local brush;
local mesh, mainShader, physicsShader, curlShader, sortShader;
local cnvs_posA, cnvs_posB, cnvs_velA, cnvs_velB;
local particleCountSQRT = 256
local particleCount = particleCountSQRT*particleCountSQRT
local debug = false
local Vector3 = require('Vector3')

function love.load()
  brush = love.graphics.newImage("particle.png")
	physicsShader = love.graphics.newShader("physics_frag.glsl")
  curlShader = love.graphics.newShader("curl_frag.glsl")
  mainShader = love.graphics.newShader("render_frag.glsl", "render_vert.glsl")
  sortShader = love.graphics.newShader("oddeven_mergesort_frag.glsl")

  -- generate the mesh data that will draw the particles out of the fbo
  local meshData = {}
  local vertMap = {}
  local numVerts = particleCount
  for i=0,(numVerts*4)-1 do
  	table.insert(meshData,{0,0, 0,0, i%4, math.floor(i/4)})
  end
  for i=0,numVerts-1 do
  	table.insert(vertMap, i*4+2)
  	table.insert(vertMap, i*4+2)
  	table.insert(vertMap, i*4+1)
  	table.insert(vertMap, i*4+3)
  	table.insert(vertMap, i*4+4)
  	table.insert(vertMap, i*4+4)
  end
  mesh = love.graphics.newMesh({{"VertexPosition", "float", 2}, {"VertexTexCoord", "float", 2}, {"PointCorner", "float", 1}, {"SpriteCount", "float", 1}}, meshData, "strip")
  mesh:setVertexMap(vertMap)

  -- nil these out so they can be gc'd
  meshData = nil
  vertMap = nil

  -- generate the fbos
  cnvs_posA = createCanvas()
  cnvs_posB = createCanvas()
  cnvs_velA = createCanvas()
  cnvs_velB = createCanvas()
  cnvs_curl = createCanvas(64,64, "linear")
  cnvs_curl:setWrap("repeat", "repeat")
  if not debug then
    -- success = love.window.setMode( 0,0, {fullscreen=true} )
  end

end
function createCanvas(x,y,scale)
	local c = love.graphics.newCanvas(x or particleCountSQRT,y or particleCountSQRT,"rgba16f")
	c:setFilter(scale or "nearest", scale or "nearest")
	return c
end
function getIdentityMat4()
  return {
    {1,0,0,0},
    {0,1,0,0},
    {0,0,1,0},
    {0,0,0,1},
  }
end
function mat4RotationY(rads)
  local mat = getIdentityMat4()
  mat[1][1] = math.cos(rads)
  mat[1][3] = -math.sin(rads)
  mat[3][1] = math.sin(rads)
  mat[3][3] = math.cos(rads)
  return mat
end
function getPerspectiveProjection(fovY, aspect, zNear, zFar)
  local r = math.rad(fovY / 2);
  local deltaZ = zFar - zNear;
  local s = math.sin(r);
  local cotangent = 0;

  if deltaZ == 0 or s == 0 or aspect == 0 then
    return nil;
  end
  cotangent = math.cos(r) / s;

  local pOut = getIdentityMat4();
  pOut[1][1] = cotangent / aspect;
  pOut[2][2] = cotangent;
  pOut[3][3] = -(zFar + zNear) / deltaZ;
  pOut[3][4] = -1;
  pOut[4][3] = -2 * zNear * zFar / deltaZ;
  pOut[4][4] = 0;

  return pOut;
end
-- returns a camera matrix looking at a point
function lookAt(eyePos, center, up)
  local f = center:copy():sub(eyePos):normalize()
  local s = f:copy():cross(up):normalize()
  local u = s:copy():cross(f)

  return {
    {s.x, u.x, -f.x, 0},
    {s.y, u.y,-f.y, 0},
    {s.z, u.z, -f.z, 0},
    {-s:dot(eyePos), -u:dot(eyePos), f:dot(eyePos), 1}
  }
end
function inverseMat4(mat)
    -- copy mat into a normal thing
    local tempMat = {
      [0] = mat[1][1], mat[1][2], mat[1][3], mat[1][4],
      mat[2][1], mat[2][2], mat[2][3], mat[2][4],
      mat[3][1], mat[3][2], mat[3][3], mat[3][4],
      mat[4][1], mat[4][2], mat[4][3], mat[4][4]
    }
    local tempTwo = {}
    tempTwo[0] = tempMat[5]  * tempMat[10] * tempMat[15] -
             tempMat[5]  * tempMat[11] * tempMat[14] -
             tempMat[9]  * tempMat[6]  * tempMat[15] +
             tempMat[9]  * tempMat[7]  * tempMat[14] +
             tempMat[13] * tempMat[6]  * tempMat[11] -
             tempMat[13] * tempMat[7]  * tempMat[10];

    tempTwo[4] = -tempMat[4]  * tempMat[10] * tempMat[15] +
              tempMat[4]  * tempMat[11] * tempMat[14] +
              tempMat[8]  * tempMat[6]  * tempMat[15] -
              tempMat[8]  * tempMat[7]  * tempMat[14] -
              tempMat[12] * tempMat[6]  * tempMat[11] +
              tempMat[12] * tempMat[7]  * tempMat[10];

    tempTwo[8] = tempMat[4]  * tempMat[9] * tempMat[15] -
             tempMat[4]  * tempMat[11] * tempMat[13] -
             tempMat[8]  * tempMat[5] * tempMat[15] +
             tempMat[8]  * tempMat[7] * tempMat[13] +
             tempMat[12] * tempMat[5] * tempMat[11] -
             tempMat[12] * tempMat[7] * tempMat[9];

    tempTwo[12] = -tempMat[4]  * tempMat[9] * tempMat[14] +
               tempMat[4]  * tempMat[10] * tempMat[13] +
               tempMat[8]  * tempMat[5] * tempMat[14] -
               tempMat[8]  * tempMat[6] * tempMat[13] -
               tempMat[12] * tempMat[5] * tempMat[10] +
               tempMat[12] * tempMat[6] * tempMat[9];

    tempTwo[1] = -tempMat[1]  * tempMat[10] * tempMat[15] +
              tempMat[1]  * tempMat[11] * tempMat[14] +
              tempMat[9]  * tempMat[2] * tempMat[15] -
              tempMat[9]  * tempMat[3] * tempMat[14] -
              tempMat[13] * tempMat[2] * tempMat[11] +
              tempMat[13] * tempMat[3] * tempMat[10];

    tempTwo[5] = tempMat[0]  * tempMat[10] * tempMat[15] -
             tempMat[0]  * tempMat[11] * tempMat[14] -
             tempMat[8]  * tempMat[2] * tempMat[15] +
             tempMat[8]  * tempMat[3] * tempMat[14] +
             tempMat[12] * tempMat[2] * tempMat[11] -
             tempMat[12] * tempMat[3] * tempMat[10];

    tempTwo[9] = -tempMat[0]  * tempMat[9] * tempMat[15] +
              tempMat[0]  * tempMat[11] * tempMat[13] +
              tempMat[8]  * tempMat[1] * tempMat[15] -
              tempMat[8]  * tempMat[3] * tempMat[13] -
              tempMat[12] * tempMat[1] * tempMat[11] +
              tempMat[12] * tempMat[3] * tempMat[9];

    tempTwo[13] = tempMat[0]  * tempMat[9] * tempMat[14] -
              tempMat[0]  * tempMat[10] * tempMat[13] -
              tempMat[8]  * tempMat[1] * tempMat[14] +
              tempMat[8]  * tempMat[2] * tempMat[13] +
              tempMat[12] * tempMat[1] * tempMat[10] -
              tempMat[12] * tempMat[2] * tempMat[9];

    tempTwo[2] = tempMat[1]  * tempMat[6] * tempMat[15] -
             tempMat[1]  * tempMat[7] * tempMat[14] -
             tempMat[5]  * tempMat[2] * tempMat[15] +
             tempMat[5]  * tempMat[3] * tempMat[14] +
             tempMat[13] * tempMat[2] * tempMat[7] -
             tempMat[13] * tempMat[3] * tempMat[6];

    tempTwo[6] = -tempMat[0]  * tempMat[6] * tempMat[15] +
              tempMat[0]  * tempMat[7] * tempMat[14] +
              tempMat[4]  * tempMat[2] * tempMat[15] -
              tempMat[4]  * tempMat[3] * tempMat[14] -
              tempMat[12] * tempMat[2] * tempMat[7] +
              tempMat[12] * tempMat[3] * tempMat[6];

    tempTwo[10] = tempMat[0]  * tempMat[5] * tempMat[15] -
              tempMat[0]  * tempMat[7] * tempMat[13] -
              tempMat[4]  * tempMat[1] * tempMat[15] +
              tempMat[4]  * tempMat[3] * tempMat[13] +
              tempMat[12] * tempMat[1] * tempMat[7] -
              tempMat[12] * tempMat[3] * tempMat[5];

    tempTwo[14] = -tempMat[0]  * tempMat[5] * tempMat[14] +
               tempMat[0]  * tempMat[6] * tempMat[13] +
               tempMat[4]  * tempMat[1] * tempMat[14] -
               tempMat[4]  * tempMat[2] * tempMat[13] -
               tempMat[12] * tempMat[1] * tempMat[6] +
               tempMat[12] * tempMat[2] * tempMat[5];

    tempTwo[3] = -tempMat[1] * tempMat[6] * tempMat[11] +
              tempMat[1] * tempMat[7] * tempMat[10] +
              tempMat[5] * tempMat[2] * tempMat[11] -
              tempMat[5] * tempMat[3] * tempMat[10] -
              tempMat[9] * tempMat[2] * tempMat[7] +
              tempMat[9] * tempMat[3] * tempMat[6];

    tempTwo[7] = tempMat[0] * tempMat[6] * tempMat[11] -
             tempMat[0] * tempMat[7] * tempMat[10] -
             tempMat[4] * tempMat[2] * tempMat[11] +
             tempMat[4] * tempMat[3] * tempMat[10] +
             tempMat[8] * tempMat[2] * tempMat[7] -
             tempMat[8] * tempMat[3] * tempMat[6];

    tempTwo[11] = -tempMat[0] * tempMat[5] * tempMat[11] +
               tempMat[0] * tempMat[7] * tempMat[9] +
               tempMat[4] * tempMat[1] * tempMat[11] -
               tempMat[4] * tempMat[3] * tempMat[9] -
               tempMat[8] * tempMat[1] * tempMat[7] +
               tempMat[8] * tempMat[3] * tempMat[5];

    tempTwo[15] = tempMat[0] * tempMat[5] * tempMat[10] -
              tempMat[0] * tempMat[6] * tempMat[9] -
              tempMat[4] * tempMat[1] * tempMat[10] +
              tempMat[4] * tempMat[2] * tempMat[9] +
              tempMat[8] * tempMat[1] * tempMat[6] -
              tempMat[8] * tempMat[2] * tempMat[5];

    local det = tempMat[0] * tempTwo[0] + tempMat[1] * tempTwo[4] + tempMat[2] * tempTwo[8] + tempMat[3] * tempTwo[12];

    if det == 0 then
      print("can't inverse matrix, determinant is zero")
        return nil
    end

    det = 1.0 / det;

    return {
      {tempTwo[0]*det, tempTwo[1]*det, tempTwo[2]*det, tempTwo[3]*det},
      {tempTwo[4]*det, tempTwo[5]*det, tempTwo[6]*det, tempTwo[7]*det},
      {tempTwo[8]*det, tempTwo[9]*det, tempTwo[10]*det, tempTwo[11]*det},
      {tempTwo[12]*det, tempTwo[13]*det, tempTwo[14]*det, tempTwo[15]*det}
    }
end
local drawCount = 0
local sortCount = 211
function love.draw()
  -- draw the curl noise into a canvas
  love.graphics.setCanvas(cnvs_curl)
  love.graphics.setShader(curlShader)
  curlShader:send("count", drawCount)
  love.graphics.draw(brush, 0,0)

  -- setup the physics drawing pass
  -- need to use ping pong fbos, because fbos cant draw into themselves
	cnvs_ping, cnvs_pong, cnvs_ping_v, cnvs_pong_v = cnvs_posA, cnvs_posB, cnvs_velA, cnvs_velB
	if drawCount%2 == 1 then
		cnvs_ping, cnvs_pong, cnvs_ping_v, cnvs_pong_v = cnvs_posB, cnvs_posA, cnvs_velB, cnvs_velA
	end
  love.graphics.setBlendMode("replace", "premultiplied")
	love.graphics.setCanvas(cnvs_ping, cnvs_ping_v)
	love.graphics.setShader(physicsShader)
	physicsShader:send("spawnOffset", drawCount)
  -- if you want to have the mouse provide gravity, use this
	physicsShader:send("mousePos", {love.mouse.getX()/love.graphics.getWidth(), love.mouse.getY()/love.graphics.getHeight()})
	-- physicsShader:send("mousePos", {0,0})
  -- physicsShader:send("mousePos", {0.5,0.5})
	physicsShader:send("velTexture", cnvs_pong_v)
	physicsShader:send("particleCount", particleCount)
	physicsShader:send("particleCountSQRT", particleCountSQRT)
  physicsShader:send("curlNoise", cnvs_curl)
	love.graphics.draw(cnvs_pong, 0,0,0,1,1)
	drawCount = drawCount+1
  -- love.graphics.setBlendMode("alpha", "premultiplied")

  -- sorting pass on the particle positions
  -- reset the stage and pass values
  -- run through the sorting a few times
      -- stage = -1
      -- pass = -1
  for i=1,20 do
    sortCount = sortCount+1
    if sortCount > 150 then
      stage = -1
      pass = -1
      sortCount = 0;
    end
    love.graphics.setBlendMode("replace", "premultiplied")
    love.graphics.setCanvas(cnvs_pong, cnvs_pong_v)
    love.graphics.setShader(sortShader)
    pass = pass-1
    if pass<0 then
      stage = stage + 1
      pass = stage
    end
    local pstage = math.pow(2, stage);
    local ppass  = math.pow(2, pass);

    -- from the gpu gems code
    -- glUniform3fARB(oddevenMergeSort.getUniformLocation("Param1"), float(pstage+pstage), 
    --    float(ppass%pstage), float((pstage+pstage)-(ppass%pstage)-1));
    -- glUniform3fARB(oddevenMergeSort.getUniformLocation("Param2"), float(width), float(height), float(ppass));
    sortShader:send("Param1", {pstage+pstage, ppass%pstage, (pstage+pstage)-(ppass%pstage)-1})
    sortShader:send("Param2", {particleCountSQRT, particleCountSQRT, ppass})
    sortShader:send("velTexture", cnvs_ping_v)
    love.graphics.draw(cnvs_ping, 0,0,0,1,1)
    -- flip/flop canvases
    local tcnvs_ping = cnvs_pong
    cnvs_pong = cnvs_ping
    cnvs_ping = tcnvs_ping
    local tcnvs_ping_v = cnvs_pong_v
    cnvs_pong_v = cnvs_ping_v
    cnvs_ping_v = tcnvs_ping_v

  end

  -- draw the main mesh
	love.graphics.setCanvas()
  love.graphics.setShader()
  love.graphics.rectangle("fill", 0,0,love.graphics.getWidth(), love.graphics.getHeight())
  love.graphics.setShader(mainShader)
  mainShader:send("posTexture", cnvs_ping)
  mainShader:send("particleCountSQRT", particleCountSQRT)
  mainShader:send("cameraViewMat", getPerspectiveProjection(70, 1, 0.001, 1000))
  -- mainShader:send("zPos", love.mouse.getX()-love.graphics.getWidth()/2)
  mainShader:send("modelMat", mat4RotationY(drawCount*0.1))
  local camDistance = 5
  local camPos = Vector3:new(math.sin(drawCount*0.01)*camDistance,0, math.cos(drawCount*0.01)*camDistance)
  camPos = Vector3:new(0.5,0,-3)
  local centerPos = Vector3:new(0,0,0)
  local up = Vector3:new(0,1,0)
  
  -- mainShader:send("cameraTransformMat",inverseMat4(lookAt(camPos, centerPos, up)))
  
  mesh:setTexture(brush)
  love.graphics.setBlendMode("alpha")
	love.graphics.draw(mesh)
	if debug then
    love.graphics.setShader()
    love.graphics.setBlendMode("alpha")
		love.graphics.draw(cnvs_ping, 0,0, 0, 1, 1)
		love.graphics.draw(cnvs_ping_v, 0,particleCountSQRT, 0, 1, 1)
    love.graphics.draw(cnvs_curl, 256, 0, 0, 1, 1)
		love.graphics.print((drawCount-1)%particleCountSQRT, 10, 10)
    love.graphics.print("Current FPS: "..tostring(love.timer.getFPS( )), 10, 30)

	end
end
-- local shaderBasic = love.graphics.newShader("basic_frag.glsl", "basic_vert.glsl")
-- local meshBounds = 0.5
-- local meshBasic = love.graphics.newMesh({
--     {-0.5, -0.5, 0,0,255,0,0,255},
--     {0.5, -0.5, 1,0,255,0,0,255},
--     {0.5,0.5,  1,1,255,0,0,255},
--     {-0.5, 0.5, 0,1,255,0,0,255}
--   }, "fan")

-- testing projection matrixes
-- function love.draw()
--   love.graphics.setShader()
--   love.graphics.rectangle("fill", 0,0,love.graphics.getWidth(), love.graphics.getHeight())
--   love.graphics.setShader(shaderBasic)
--   shaderBasic:send("cameraViewMat", getPerspectiveProjection(90, 1, 0.001, 1000))
--   shaderBasic:send("modelMat", mat4RotationY(drawCount*0.01))
--   -- shaderBasic:send("zPos", love.mouse.getX()-love.graphics.getWidth()/2)
--   love.graphics.draw(meshBasic, 0,0)
--   drawCount = drawCount+1
--   -- love.graphics.draw(brush, 0,0)
-- end
function love.keypressed()
  -- love.event.push ("quit")
end

